from abc import ABCMeta

class styles(dict):
  def __init__(self, *args, **props):
    super().__init__()
    if args:
      if isinstance(args[0], dict):
        self.update(args[0])
    if isinstance(args[0], str):
      for l in args[0].split(';'):
        l = l.strip()
        if l:
          k, v = l.split(':')

          self[k.replace('-', '_').strip()] = v.strip()

    self.update(props)

  def __str__(self):
    return ' '.join(f"{k.replace('_', '-')}: {v};"
                    for k, v in self.items())


class Node:
  NAME = None
  CHILDREN = False

  def __init__(self, *children, style=None, **props):
    self._display_handle = None
    self.children = list(children)
    self.props = props
    self.style = style if isinstance(style, styles) else styles(style)

  @property
  def name(self):
    return self.NAME or self.__class__.__name__

  def __str__(self):
    return self.html()

  __repr__ = __str__

  _repr_html_ = __str__

  def html(self):
    if self.CHILDREN:
      return f"""
         <{self.name} {' '.join(
        f'{k}="{v}"' for k, v in self.props.items())}  style="{self.style}"/>
      """

    return f"""
    <{self.name} {' '.join(
      f'{k}="{v}"' for k, v in self.props.items())} style="{self.style}">
       {self.format_children()}
    </{self.name}>
    """

  def __call__(self, *children):
    self.children = list(children)
    return self

  def render(self):
    from IPython.core.display import HTML
    return HTML(self.html())

  def format_children(self):
    return ' '.join(str(c) for c in self.children)

  def display(self):
    from IPython.core.display import display
    self._display_handle = display(self.render(), display_id=True)

  def update(self):
    if self._display_handle:
      self._display_handle.update(self.render())


def prop(name):
  def g(self):
    return getattr(self, f'_{name}')

  def s(self, val):
    setattr(self, f'_{name}', val)
    self.update()

  return property(g, s)


class ReactiveNodeMeta(ABCMeta):
  def __new__(mcls, name, bases, namespace):
    annotations = namespace.get('__annotations__', {})

    props = set()
    prop_defaults = {}
    for name, annotation in annotations.items():
      if annotation is prop:
        if name in namespace:
          prop_defaults[name] = namespace[name]
        namespace[name] = prop(name)
        props.add(name)
    if '_props' in namespace:
      namespace['_props'].update(props)
    else:
      namespace['_props'] = props

    if '_default_props' in namespace:
      namespace['_default_props'].update(prop_defaults)
    else:
      namespace['_default_props'] = prop_defaults

    return super().__new__(mcls, name, bases, namespace)

class ReactiveMixin(metaclass=ReactiveNodeMeta):
  _props: set
  _default_props: dict

  def __init__(self, *args, **props):
    super(ReactiveMixin, self).__init__(*args, **props)
    self._init_props(**props)

  def _init_props(self, **passed_props):
    for prop in self._props:
      if prop not in passed_props and prop in self._default_props:
        setattr(self, prop, self._default_props[prop])
      elif prop in passed_props:
        setattr(self, prop, passed_props[prop])


class EmptyNode(Node):
  CHILDREN = True


class div(Node): pass


class span(Node): pass


class img(EmptyNode): pass
class p(Node): pass
class h1(Node): pass
class h2(Node): pass
class h3(Node): pass
class h4(Node): pass
class progress(EmptyNode): pass


class matplotfig(img):
  NAME = 'img'

  def __init__(self, figure, live=False, style=None, **props):
    super(matplotfig, self).__init__(style=style, **props)

    self.figure = figure
    self.live = live

    if not self.live:
      from .plot import figure_to_base64
      self.props['src'] = figure_to_base64(figure, data_encoded=True)


  def html(self):
    if self.live:
      from .plot import figure_to_base64
      self.props['src'] = figure_to_base64(self.figure, data_encoded=True)
    if 'src' not in self.props:
      from .plot import figure_to_base64
      self.props['src'] = figure_to_base64(self.figure, data_encoded=True)

    return super(matplotfig, self).html()



def _cell(val, size=25):
  return div(
    style=(
      f'width: {size}px; height: {size}px;'
      f' display: inline-block;'
      f' background-color: rgba({50 + val}, {20 + val * .8}, {80 + val * .6}, 1);'
      f' margin: 0; border: 1px solid rgba(0,0,0, .05)'
    ))


def _row(*args):
  return div(*args, style='margin: 0; padding: 0; font-size:0; white-space: nowrap;')


def tensor(t, cell_size=15, scaled=None, min=None, max=None, max_elements=50000):
  if t.numel() > max_elements:
    raise ValueError('tensor too large')
  if scaled is None:
    max = t.max() if max is None else max
    min = t.min() if min is None else min
    scaled = (t.float() - min) / (max-min) * 255

  if t.ndim == 1:
    return div(
      f"shape: {tuple(t.shape)}",
      _row(*(_cell(v, size=cell_size) for v in scaled))
    )

  if t.ndim == 2:
    return div(
      f"shape: {tuple(t.shape)}",
      *[
        _row(*(_cell(v, size=cell_size) for v in row)) for row in scaled
      ]
    )

  if t.ndim >= 3:
    return div(
      f"shape: {tuple(t.shape)}",
      div(
        *(div(style='border: 1px solid #CCC; padding: 5px; margin: 3px; display: inline-block; border-radius: 6px')(
          f"index: {n}", tensor(x, scaled=x, cell_size=cell_size, min=min, max=max)
        ) for n, x in enumerate(scaled)
        )
      )
    )


def zero_out_nans(module, grad_input, grad_output):
  """
  Backwards hook to turn NaNs in gradients to 0
  """
  for grad in grad_input:
    grad[grad != grad] = 0  # technically shouldn't modify inputs



class Hook:
  def __init__(self, module):
    self.module = module
    self.handles = None
    self.register(self.module)

  def register(self, module):
    pass

  def remove(self):
    pass

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.remove()


def shape_hook(name, depth=None, show=False):
    depth = depth if depth is not None else name.count('.')
    indent = '  ⬐' * depth if name else ''

    def print_shapes(m, input, output):
      print(
        f'{f"{indent}{name}:": <20}',
        f"{f'({m.__class__.__name__})': <15}",
        ', '.join(f"{str(tuple(x.shape)): <15}" for x in input),
        "=>",
        f"{str(tuple(output.shape)): <15}")

      if show is not False:
        import yann
        if show is True:
          yann.show(output)
        else:
          yann.show(output[show])

    return print_shapes


class ShapeHook(Hook):
  def __init__(self, module, show=False):
    self.show = show
    super().__init__(module)


  def register(self, module):
    self.handles = []
    for n, m in module.named_modules():
      self.handles.append(m.register_forward_hook(shape_hook(n, show=self.show)))

  def remove(self):
    for h in self.handles:
      h.remove()


import torch

def evaluate_metrics(targets=None, outputs=None, metrics=None):
  values = {}
  with torch.no_grad():
    for name, metric in metrics.items():
      values[name] = metric(targets, outputs)
  return values


class Evaluator:
  def update(self):
    pass

  def metrics(self):
    pass

  def report(self):
    pass

  def summary(self):
    pass

  def help(self):
    pass

  def show(self):
    pass


class MultiClassEvaluator(Evaluator):
  metrics = {}
  pass


class BinaryClassificationEvaluator(Evaluator):
  pass


class RankingEvaluator(Evaluator):
  pass


class RegressionEvaluator(Evaluator):
  pass


class RetrievalEvaluator(Evaluator):
  pass


from .html import div, prop, ReactiveMixin, Node


class ProgressBar(ReactiveMixin, Node):
  value: prop = 0
  max: prop = 100
  color: prop = 'lightgreen'
  background: prop = 'lightgray'

  def html(self):
    return (
      div(
        style=f"""
        position: relative;
        min-height: 20px; 
        border-radius: 6px; 
        background-color: {self.background}; 
        overflow: hidden; 
        padding: 5px 10px;
        """
      )(
        div(style='position:relative; z-index: 1;')(
          *self.children
        ),
        div(
          style=f"""
          background-color: {self.color};
          position: absolute;
          top: 0;
          bottom: 0;
          left: 0;
          right: {100 - self.value / self.max * 100}%;
          z-index: 0;
        """)
        )
    ).html()


import re
import sys

import numpy as np
import torch
from PIL import Image
import datetime


def timestr(d=None):
  return f"{(d or datetime.datetime.utcnow()).strftime('%y-%m-%dT%H:%M:%S')}"


def camel_to_snake(text):
  s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
  return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def abbreviate(text):
  return re.sub(r"([a-zA-Z])[a-z]*[^A-Za-z]*",r"\1", text).lower()


def get_arg_parser(x, description=None, epilog=None, parser=None, **kwargs):
  import argparse
  from ..params import Field
  parser = parser or argparse.ArgumentParser(description=description, epilog=epilog, **kwargs)

  abbreviations = set()

  for k, v in x.items():
    names = []
    abbreviated = abbreviate(k)
    if abbreviated not in abbreviations:
      names.append(f"-{abbreviated}")
      abbreviations.add(abbreviated)
    names.append(f"--{camel_to_snake(k)}")

    if isinstance(v, dict):
      parser.add_argument(
        *names,
        default=v.get('default'),
        type=v.get('type'),
        action=v.get('action'),
        help=v.get('help'),
        required=v.get('required'),
        choices=v.get('choices'),
        dest=v.get('dest')
      )
    elif isinstance(v, Field):
      parser.add_argument(
        *names,
        default=v.default,
        type=v.type,
        help=f"{v.help or k} (default: {v.default})",
        required=v.required,
        choices=getattr(v, 'choices', None),
      )
    else:
      parser.add_argument(*names, default=v, type=type(v))
  return parser


def truthy(items):
  return [x for x in items if x]


class Obj(dict):
  __setattr__ = dict.__setitem__
  __getattr__ = dict.__getitem__


def almost_equal(t1, t2, prec=1e-12):
  return torch.all(torch.lt(torch.abs(torch.add(t1, -t2)), prec))


def equal(t1, t2):
  return torch.all(t1 == t2)


def randimg(*shape, dtype='uint8'):
  return Image.fromarray((np.random.rand(*shape) * 255).astype(dtype))


def progress(it, num=None):
  if not num:
    try:
      num = len(it)
    except:
      num = None

  if num:
    for n, x in enumerate(it, 1):
      sys.stdout.write(f"\r{n} / {num}")
      sys.stdout.flush()
      yield x
  else:
    for n, x in enumerate(it, 1):
      sys.stdout.write(f"\r{n}")
      sys.stdout.flush()
      yield x


def repeat(val):
  while True:
    yield val


def counter(start=0, end=None, step=1):
  current = start
  while end is None or (end and (current < end)):
    yield current
    current += step


def to_numpy(x):
  if isinstance(x, np.ndarray):
    return x
  if torch.is_tensor(x):
    return x.to('cpu').detach().numpy()
  return np.array(x)


class RangeMap(dict):
  def __init__(self, items=None):
    super().__init__()

    if isinstance(items, dict):
      items = items.items()

    if items:
      for k, v in items:
        self[k] = v

  def __getitem__(self, item):
    if isinstance(item, tuple):
      return super().__getitem__(item)

    if item is not None:
      for (min, max), value in self.items():
        if min is not None and item < min:
          continue
        if max is not None and item > max:
          continue
        return value

    raise KeyError(f'Key `{item}` does not fall in any of the ranges')

  def __call__(self, item):
    return self[item]


def pretty_size(bytes):
  num = bytes
  for unit in ['bytes', 'KB', 'MB', 'GB', 'TB']:
    if num < 1024.0:
      return f'{num:3.1f} {unit}'
    num /= 1024.0


def print_tree(root, indent=2, depth=None, filter=None):
  from pathlib import Path
  from datetime import datetime
  root = Path(root)
  for path in sorted((root, *root.rglob('*'))):
    d = len(path.relative_to(root).parts)
    if depth and depth < d:
      continue
    if not path.is_dir() and filter and not path.match(filter):
      continue
    if path.is_dir():
      print(f'{" " * (d * indent)} /{path.name}/')
    else:
      print(
        f'{" " * (d * indent)}  - {path.name:25} '
        f'{f"({pretty_size(path.stat().st_size)})":15} '
        f'{datetime.fromtimestamp(path.stat().st_mtime)}'
      )


def fully_qualified_name(x):
  module = x.__class__.__module__
  if module is None or module == str.__class__.__module__:
    return x.__class__.__name__
  else:
    return f'{module}.{x.__class__.__name__}'


def hash_params(module):
    from hashlib import sha1
    s = sha1()
    for p in module.parameters():
      s.update(to_numpy(p).tobytes())
    return s.hexdigest()


from torch import nn

from .conv import (
  ConvBlock,
  ConvBlock1x1,
  ConvBlock3x3,
  MixConv,
  SqueezeExcitation,
  EfficientChannelAttention,
  DepthwiseSeparableConv2d,
  DepthwiseConv2d
)
from .shape import View, Flatten, Infer, Squeeze, Reshape, Permute, Transpose, FlattenSequences
from .stack import Stack
from .residual import Residual


class Init:
  def __init__(self, cls, *args, **kwargs):
    self.cls = cls
    self.args = args
    self.kwargs = kwargs

  def __call__(self, *args, **kwargs):
    return self.cls(*self.args, *args, **{**self.kwargs, **kwargs})



class TrainEvalSwitch(nn.Module):
  def __init__(self, train=None, eval=None):
    super().__init__()
    self.train = train
    self.eval = eval

  def forward(self, *args):
    if self.training:
      return self.train(*args) if self.train else args
    else:
      return self.eval(*args) if self.eval else args



import lmdb
import pickle
import json
from contextlib import contextmanager

from ..serialize import serialize_arrow, deserialize_arrow, to_bytes, to_unicode
from ..images import image_to_bytes, image_from_bytes


class LMDB:
  def __init__(self, path, map_size=1e10, **kwargs):
    self.path = path
    self.db = None
    self.open(map_size=map_size, **kwargs)

    self._current_transaction = None

  @property
  def stats(self):
    return self.db.stat()

  def __len__(self):
    return self.stats['entries']

  def close(self):
    self.db.close()

  def open(self, **kwargs):
    self.db = lmdb.open(self.path, **kwargs)

  def __del__(self):
    self.close()
    del self.db

  def __getitem__(self, key):
    # TODO: support indexing multiple values
    if self._current_transaction:
      return self.deserialize(self._current_transaction.get(self.serialize_key(key)))
    else:
      with self.db.begin(write=False) as t:
        return self.deserialize(t.get(self.serialize_key(key)))

  def __setitem__(self, key, value):
    if self._current_transaction:
      return self._current_transaction.put(self.serialize_key(key), self.serialize(value))
    else:
      with self.db.begin(write=True) as t:
        return t.put(self.serialize_key(key), self.serialize(value))

  def __delitem__(self, key):
    if self._current_transaction:
      return self._current_transaction.delete(self.serialize_key(key))
    else:
      with self.db.begin(write=True) as t:
        return t.delete(self.serialize_key(key))

  def __iter__(self):
    if self._current_transaction:
      for k, v in self._current_transaction.cursor():
        yield self.deserialize_key(k), self.deserialize(v)
    else:
      with self.db.begin(write=False) as t:
        for k, v in t.cursor():
          yield self.deserialize_key(k), self.deserialize(v)

  def update(self, items):
    if isinstance(items, dict):
      items = items.items()
    if self._current_transaction:
      for k, v in items:
        self._current_transaction.put(self.serialize_key(k), self.serialize(v))
    else:
      with self.db.begin(write=True) as t:
        for k, v in items:
          t.put(self.serialize_key(k), self.serialize(v))

  @contextmanager
  def transaction(self, write=False, buffers=False):
    with self.db.begin(write=write, buffers=buffers) as t:
      self._current_transaction = t

      yield

      self._current_transaction = None

  @staticmethod
  def serialize_key(x):
    return x

  @staticmethod
  def deserialize_key(x):
    return x

  @staticmethod
  def serialize(x):
    return x

  @staticmethod
  def deserialize(x):
    return x


class ArrowLMDB(LMDB):
  """
  LMDB that uses arrow to serialize the values
  """

  @staticmethod
  def serialize_key(x): 
    return to_bytes(x)

  @staticmethod
  def deserialize_key(x): 
    return to_unicode(x)

  @staticmethod
  def serialize(x): 
    return serialize_arrow(x)

  @staticmethod
  def deserialize(x): 
    return deserialize_arrow(x)


class PickleLMDB(LMDB):
  @staticmethod
  def serialize_key(x):
    return to_bytes(x)

  @staticmethod
  def deserialize_key(x):
    return to_unicode(x)

  @staticmethod
  def serialize(x):
    return pickle.dumps(x, protocol=-1)

  @staticmethod
  def deserialize(x):
    return pickle.loads(x)


class ImageLMDB(LMDB):
  format = 'jpeg'

  def serialize_key(self, x):
    return to_bytes(x)

  def deserialize_key(self, x):
    return to_unicode(x)

  def serialize(self, x):
    return image_to_bytes(x, format=self.format)

  def deserialize(self, x):
    return image_from_bytes(x)


class JSONLMDB(LMDB):
  @staticmethod
  def serialize_key(x):
    return to_bytes(x)

  @staticmethod
  def deserialize_key(x):
    return to_unicode(x)

  @staticmethod
  def serialize(x):
    return to_bytes(json.dumps(x))

  @staticmethod
  def deserialize(x):
    return json.loads(to_unicode(x))


from .plot import plot_line, plot_pred_scores, plot_rocs, \
  plot_confusion_matrix, plot_cooccurrences
from .image import show_images



class Plotter:
  def __call__(self, *args, **kwargs):
    pass

class Shower:
  def __call__(self, x, format=None, **kwargs):
    if isinstance(x, (list, tuple)):
      if isinstance(x[0], str):
        if x[0].lower().endswith(('.jpg', '.jpeg', '.png')):
          return self.images(x)

      try:
        from PIL import Image
        if isinstance(x[0], Image.Image):
          return self.images(x)
      except:
        pass
    if isinstance(x, str):
      if x.lower().endswith(('.jpg', '.jpeg', '.png')):
        return self.images(x)

    try:
      from PIL import Image
      if isinstance(x, Image.Image):
        return self.images(x)
    except:
      pass

    import torch
    if isinstance(x, torch.Tensor):
      return self.tensor(x)


    return x

  def images(self, *args, **kwargs):
    return show_images(*args, **kwargs)

  def tensor(self, t, *args, **kwargs):
    from .html import tensor
    return tensor(t, *args, **kwargs).display()

show = Shower()
plot = Plotter()